Program LongFair;

{$mode objfpc}
{.$define debug}

Uses
  SysUtils, NCalc, BoolFactory, Expressions;

type
  TDigit = '0'..'9';
  TDigitSet = set of TDigit;


  TNumberFact = class
  private
    FNumber: TValue;
  protected
    function ExtractDigits: TDigitSet;
  public
    function IsFair: OPBool;
    constructor Create(ANr: TValue);
    property Number: TValue read FNumber;
  end;

  TNumberFactLister = class
  private
    FStart: TValue;
    FStop: TValue;
  protected
  public
    procedure GetLimits;
    procedure ListFairNumbers;

  end;

const
  ListHeader = 'Listing all "fair" numbers from %s to %s';
  FoundFairNumber = 'Found a fair number: %s';
  AskLowerLimit = 'Enter value for lower limit: ';
  AskUpperLimit = 'Enter value for upper limit: ';

{$ifdef debug}
function DbgS(Digits: TDigitSet): String;
var
  D: TDigit;
begin
  Result := '[';
  for D := Low(TDigit) to High(TDigit) do
    if D in Digits then
      Result := Result + D + ',';
  if (Result[Length(Result)] = ',') then
    System.Delete(Result, Length(Result), 1);
  Result := Result + ']';
end;
{$endif}

{TNumberFact}

function TNumberFact.ExtractDigits: TDigitSet;
const
  AllDigits: TDigitSet = ['0','1','2','3','4','5','6','7','8','9'];
var
  TempNr: TValue;
  ADigit: TDigit;
begin
  Result := [];
  TempNr := AbsN(FNumber);
  while IsFalse(IsZeroN(TempNr)) and IsFalse((AllDigits * Result) = AllDigits) do
  begin
    ADigit := NumberToPrimitiveN(ModN(TempNr, '10'));
    Result := Result + [ADigit];
    TempNr := DivN(TempNr, '10');
  end;
end;

function TNumberFact.IsFair: OPBool;
var
  ADigit: TDigit;
  Digits: TDigitSet;
begin
  Result := False;
  if IsTrue(IsZeroN(FNumber)) then Exit; //shortcut
  Digits := ExtractDigits;
  {$ifdef debug}
  writeln(format('FNumber = %s, Digits = %s',[TempNr,DbgS(Digits)]));
  {$endif}
  for ADigit := Low(TDigit) to High(TDigit) do
  begin
    if ADigit in Digits then
      if IsTrue(IsZeroN(ADigit)) or IsFalse(IsZeroN(ModN(FNumber, ADigit))) then Exit; //False
  end;
  Result := True;
end;

constructor TNumberFact.Create(ANr: TValue);
begin
  AssertIsNumberN(ANr);
  FNumber := ANr;
end;



{TNumberFactLister}

procedure TNumberFactLister.ListFairNumbers;
var
  i: TValue;
  NumberFact: TNumberFact;
begin
  writeln(Format(ListHeader,[FStart, FStop]));
  i := FStart;
  repeat
    NumberFact := TNumberFact.Create(i);
    try
      if NumberFact.IsFair then writeln(Format(FoundFairNumber,[i]));
    finally
      NumberFact.Free;
    end;
    i := AddN(i, '1');
  until IsTrue(MoreN(i, FStop));
end;

procedure TNumberFactLister.GetLimits;
//either accept commandline parameters or ask for the limits
var
  HaveStart, HaveStop: OPBool;
  Temp: TValue;
begin
  {$ifdef debug}
  FStart := '10';
  FStop := '256';
  Exit;
  {$endif debug}
  HaveStart := OPBool(False);
  HaveStop := OPBool(False);
  if (ParamCount > 0) and IsTrue(IsNumberN(ParamStr(1))) then
  begin
    HaveStart := OPBool(True);
    FStart := ParamStr(1);
    if (ParamCount > 1) and IsTrue(IsNumberN(ParamStr(2))) then
    begin
      HaveStop := OPBool(True);
      FStop := ParamStr(2);
    end;
  end;
  while IsFalse(HaveStart) do
  begin
    write(AskLowerLimit);
    readln(FStart);
    HaveStart := IsTrue(IsNumberN(FStart));
  end;
  while not HaveStop do
  begin
    write(AskUpperLimit);
    readln(FStop);
    HaveStop := IsTrue(IsNumberN(FStop));
  end;
  if IsTrue(MoreN(FStart, FStop)) then
  begin
    Temp := FStart;
    FStart := FStop;
    FStop := Temp;
  end;
end;

var
  NumberFactLister: TNumberFactLister;

begin
  NumberFactLister := TNumberFactLister.Create;
  try
    NumberFactLister.GetLimits;
    NumberFactLister.ListFairNumbers;
  finally
    NumberFactLister.Free;
  end;
end.

